import System.Environment
import System.Console.GetOpt
import System.Process
import GHC.IO.Exception( ExitCode( ExitSuccess ) )
import System.Directory

import Text.Parsec.Error
import Data.Array.Unboxed
import Data.List.Split
import Data.Word
import Data.Tuple (swap)
import qualified Data.ByteString as B
import qualified Data.ByteString.Lazy as L
import qualified Data.Vector as V
--import Control.Arrow
import Control.Monad

import qualified Codec.Picture as J
import qualified Graphics.Pgm as P


mapTuple :: (a -> b) -> (a, a) -> (b, b)
mapTuple f (a1, a2) = (f a1, f a2)

getNum :: Either e (a t Int) -> Either String (a t Int)
getNum (Left x) = Left "NG"
getNum (Right x)  = return x
--getPgm :: Either err StorageImage -> Either String Grey 
--getPgm (Left err) = Left "Error while reading the image."
--getPgm (Right x) = Right $ convert x 

--toImg :: Either String J.DynamicImage -> Either String (J.Image J.PixelRGB8)
--toImg (Left b) = Left b
--toImg (Right (J.ImageRGB8 b) ) = Right b

data Color = Red | Green | Blue deriving (Show, Eq)
type Row = V.Vector Color
type FilterImage = V.Vector Row
type Length = Int
type Filter = (Length,FilterImage)

toColor :: Char -> Color
toColor 'R' = Red
toColor 'G' = Green
toColor 'B' = Blue

tostrRow :: String -> (Length, [String])
tostrRow str
    | l `elem` [x^2 | x <- [2,3..]] = (r,f)
    | otherwise = error "FilterString can not devide equally"
    where l = length str
          r = floor $ sqrt $ fromIntegral l
          f = flip chunksOf str r

toFilter :: String -> Int -> Int -> Filter
toFilter str x y = (l,f)
    where (l,s) = tostrRow str
          f = V.fromListN x $ cycle $ map (\xs -> V.fromListN y $ cycle $ map toColor xs) s

toRGBPixel :: Word16 -> Color -> J.PixelRGB16
toRGBPixel x Red   = J.PixelRGB16 x 0 0
toRGBPixel x Green = J.PixelRGB16 0 x 0
toRGBPixel x Blue  = J.PixelRGB16 0 0 x

toWord16 :: Int -> Word16
toWord16 x = fromIntegral x :: Word16  


selfprod xs = [(x,y) | x <-xs, y <- xs]

splitbyColor :: V.Vector (Int, Int) -> FilterImage -> Color -> V.Vector (Int, Int)
splitbyColor n f c = V.filter select n
            where (!) = (V.!)
                  select (x, y) = c == f!y!x

mapN f nl = map (\l -> V.map f l) nl

eqval :: V.Vector Int -> Word16
eqval xs
    | xs == V.empty = error "not contain color"
    | otherwise = toWord16 $ (V.sum xs) `div` (V.length xs)

tplus (a,b) (c,d) = (a+c , b+d)

colors = [Red,Green,Blue]
pixelReduce :: PGM -> Filter -> V.Vector (Int,Int) -> Int -> Int -> J.PixelRGB16
pixelReduce p (l,i) m x y = J.PixelRGB16 r g b
      where o = (y*l,x*l)
            --m = V.fromList $ selfprod [0,1..(l-1)]
            targets = V.map (tplus o) m
            [r,g,b] = map eqval $ mapN (p!) $ flip map colors $ splitbyColor targets i
            --r = eqval $ map (p!) $ splitbyColor targets f Red
            --g = eqval $ map (p!) $ splitbyColor targets f Green
            --b = eqval $ map (p!) $ splitbyColor targets f Blue

type PGM = UArray (Int,Int) Int
toImage :: String -> Either ParseError PGM -> J.Image J.PixelRGB16
toImage s (Left e) = error $ "ignore convert because: " ++ (show e)
toImage s (Right p) = J.generateImage colgen width height
    where (h,w) = mapTuple (+1) $ snd $ bounds p
          (l,f) = toFilter s w h
          (width, height) = mapTuple (flip div l) (w,h)
          m = V.fromList $ selfprod [0,1..(l - 1)]
          colgen x y = pixelReduce p (l,f) m x y
    --where (width, height) = mapTuple (+1) $ snd $ bounds p
        　　--colZip col ((x,y),pix) =  (toRGBPixel . toWord16) pix
        　　--colgen x y =  (toRGBPixel . toWord16) (p!(x,y)) ((snd f)!!y!!x)
          



data Flag = Help | Fi String | Output String deriving (Eq, Show)

progName :: String
progName = "reducedevelop"

options :: [OptDescr Flag]
options =
  [ Option ['h','?'] ["help"] (NoArg Help) "display this help."
  , Option ['f'] ["filter"] (ReqArg Fi "Filter")  "write RGB filter. like '-f BGGR' '-f RGGBRGGBRGGBRGGB' default is 'RGGB'"
  --, Option ['i'] ["input"] (ReqArg Input "FILE")  "input file"
  --, Option ['o'] ["output"] (OptArg poutput "FILE")  "output file"
  ]

poutput :: Maybe String -> Flag
poutput Nothing = Output "output.png"
poutput (Just s) = Output s


getOpts :: [String] -> IO ([Flag], [String])
getOpts argv = 
  case getOpt RequireOrder options argv of
     ((Help :xs),_,_) -> error $ usageInfo header options
     (o,n,[]  ) -> return (o,n)
     (_,_,errs) -> ioError (userError (concat errs ++ usageInfo header options))
  where header = "Usage: wildraw [OPTION...] INPUTFILE [OUTPUTFILE]"

parseArgs :: [String] -> (String,String)
parseArgs (i:o:xs) = (i,o)
parseArgs (i:[]) = (i,"output.png")
parseArgs [] = error "no input file"

takeL x = reverse . take x . reverse

callpgm :: String -> Bool
callpgm x = c == ".pgm" || c == ".PGM"
        where c = takeL 4 x

pinput str
    |callpgm str = str
    |otherwise = "tmp.pgm"

fromOption :: Flag -> (String,String)
fromOption (Fi f) = ("fi",f)
fromOption _ = ("","")

filterget :: Maybe String -> String
filterget Nothing = "RGGB"
filterget (Just a) = a

--filterget :: [a] -> String
--filterget [] = "RGGB"
--filterget (Fi x) : xs = x
--filterget xs = filterget xs 

main = do

    (opts, args) <- getOpts =<< getArgs
    --print (opts,args)

    let (inputfile,outputfile) = parseArgs args
    let i = pinput inputfile
    if callpgm inputfile
        then return ExitSuccess
        else system $ "dcraw -D -6 -W -c " ++ inputfile ++ " >tmp.pgm"

    bytes <- B.readFile i

    bytes <- B.readFile "tmp.pgm"
    let res = P.pgmToArray bytes

    let optd = map fromOption opts
    let fs = filterget $ lookup "fi" optd
    let image = toImage fs res


    J.writePng outputfile image
    --J.writePng "out.png" image

